/*
@file    FT800_commands.c
@brief   Contains Functions for using the FT8xx
@version 2.4
@date    2016-07-26
@author  Rudolph Riedel

This file needs to be renamed to FT800_command.cpp for use with Arduino. 

@section History

2.1
- removed the REG_ROTATE line from ft800_init()
- simplified ft800_busy() slightly
- implemented ft800_cmd_mediafifo(), ft800_cmd_setrotate(), ft800_cmd_setbitmap() (FT81x)

2.2
- moved hardware abstraction functions over to FT800_config.c

2.3
- changes to this header

2.3
- added "ft800_cmdWrite(FT_CLKINT);" to top of ft800_init() to switch easier
  between displays with and without crystal populated
  note: need to add a define for internal/external clock to

- added a ft800_get_cmdoffset() call to the end of ft800_init() as it must be called
  after init anyways

*/

#include "FT800.h"
#include "FT800_config.h"


// FT800 Memory Commands - use with ft800memWritexx and ft800memReadxx
#define MEM_WRITE	0x80	// FT800 Host Memory Write
#define MEM_READ	0x00	// FT800 Host Memory Read

volatile uint16_t cmdOffset = 0x0000;		// used to navigate command ring buffer


void ft800_cmdWrite(uint8_t data)
{
	ft800_cs_set();
	spi_transmit(data);
	spi_transmit(0x00);
	spi_transmit(0x00);
	ft800_cs_clear();
}


uint8_t ft800_memRead8(uint32_t ftAddress)
{
	uint8_t ftData8 = 0;
	ft800_cs_set();
	spi_transmit((uint8_t)(ftAddress >> 16) | MEM_READ); // Send Memory Write plus high address byte
	spi_transmit((uint8_t)(ftAddress >> 8));		// Send middle address byte
	spi_transmit((uint8_t)(ftAddress));		// Send low address byte
	spi_transmit(0x00);				// Send dummy byte
	ftData8 = spi_receive(ftData8);		// Read data byte
	ft800_cs_clear();
	return ftData8;				// Return byte read
}


uint16_t ft800_memRead16(uint32_t ftAddress)
{
	uint16_t ftData16 = 0;
	ft800_cs_set();
	spi_transmit((uint8_t)(ftAddress >> 16) | MEM_READ); // Send Memory Write plus high address byte
	spi_transmit((uint8_t)(ftAddress >> 8));		// Send middle address byte
	spi_transmit((uint8_t)(ftAddress));		// Send low address byte
	spi_transmit(0x00);				// Send dummy byte
	ftData16 = (spi_receive(0x00));		// Read low byte
	ftData16 = (spi_receive(0x00) << 8) | ftData16; // Read high byte
	ft800_cs_clear();
	return ftData16;				// Return integer read
}


uint32_t ft800_memRead32(uint32_t ftAddress)
{
	uint32_t ftData32= 0;
	ft800_cs_set();
	spi_transmit((uint8_t)(ftAddress >> 16) | MEM_READ); // Send Memory Write plus high address byte
	spi_transmit((uint8_t)(ftAddress >> 8));		// Send middle address byte
	spi_transmit((uint8_t)(ftAddress));		// Send low address byte
	spi_transmit(0x00);				// Send dummy byte
	ftData32 = ((uint32_t)spi_receive(0x00));		// Read low byte
	ftData32 = ((uint32_t)spi_receive(0x00) << 8) | ftData32;
	ftData32 = ((uint32_t)spi_receive(0x00) << 16) | ftData32;
	ftData32 = ((uint32_t)spi_receive(0x00) << 24) | ftData32; // Read high byte
	ft800_cs_clear();
	return ftData32;				// Return long read
}


void ft800_memWrite8(uint32_t ftAddress, uint8_t ftData8)
{
	ft800_cs_set();
	spi_transmit((uint8_t)(ftAddress >> 16) | MEM_WRITE); // Send Memory Write plus high address byte
	spi_transmit((uint8_t)(ftAddress >> 8));		// Send middle address byte
	spi_transmit((uint8_t)(ftAddress));		// Send low address byte
	spi_transmit(ftData8);			// Send data byte
	ft800_cs_clear();
}


void ft800_memWrite16(uint32_t ftAddress, uint16_t ftData16)
{
	ft800_cs_set();
	spi_transmit((uint8_t)(ftAddress >> 16) | MEM_WRITE); // Send Memory Write plus high address byte
	spi_transmit((uint8_t)(ftAddress >> 8));		// Send middle address byte
	spi_transmit((uint8_t)(ftAddress));		// Send low address byte
	spi_transmit((uint8_t)(ftData16));		// Send data low byte
	spi_transmit((uint8_t)(ftData16 >> 8));		// Send data high byte
	ft800_cs_clear();
}


void ft800_memWrite32(uint32_t ftAddress, uint32_t ftData32)
{
	ft800_cs_set();
	spi_transmit((uint8_t)(ftAddress >> 16) | MEM_WRITE); // Send Memory Write plus high address byte
	spi_transmit((uint8_t)(ftAddress >> 8));		// Send middle address byte
	spi_transmit((uint8_t)(ftAddress));		// Send low address byte
	spi_transmit((uint8_t)(ftData32));		// Send data low byte
	spi_transmit((uint8_t)(ftData32 >> 8));
	spi_transmit((uint8_t)(ftData32 >> 16));
	spi_transmit((uint8_t)(ftData32 >> 24));		// Send data high byte
	ft800_cs_clear();
}


// Check if the graphics processor completed executing the current command list.
// This ist the case when REG_CMD_READ matches cmdOffset, indicating that all commands have been executed.
uint8_t ft800_busy(void)
{
	uint16_t cmdBufferRead;

	cmdBufferRead = ft800_memRead16(REG_CMD_READ);	// Read the graphics processor read pointer

	if(cmdOffset != cmdBufferRead)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}


uint32_t ft800_get_touch_tag(void)
{
	uint32_t value;

	value = ft800_memRead32(REG_TOUCH_TAG);
	return value;
}


// Order the command co-prozessor to start processing its FIFO que
void ft800_cmd_execute(void)
{
	uint32_t ftAddress;

	ftAddress = REG_CMD_WRITE;

	ft800_cs_set();
	spi_transmit((uint8_t)(ftAddress >> 16) | MEM_WRITE); // Send Memory Write plus high address byte
	spi_transmit((uint8_t)(ftAddress >> 8));	// Send middle address byte
	spi_transmit((uint8_t)(ftAddress));			// Send low address byte
	spi_transmit((uint8_t)(cmdOffset));			// Send data low byte
	spi_transmit((uint8_t)(cmdOffset >> 8));	// Send data high byte
	ft800_cs_clear();
}


void ft800_get_cmdoffset(void)
{
	cmdOffset = ft800_memRead16(REG_CMD_WRITE);
}


void ft800_inc_cmdoffset(uint16_t increment)
{
	cmdOffset += increment;
	cmdOffset &= 0x0fff;
}


// Beginn a co-prozessor command
void ft800_start_cmd(uint32_t command)
{
	uint32_t ftAddress;

	ftAddress = FT_RAM_CMD + cmdOffset;
	ft800_cs_set();
	spi_transmit((uint8_t)(ftAddress >> 16) | MEM_WRITE); // Send Memory Write plus high address byte
	spi_transmit((uint8_t)(ftAddress >> 8));	// Send middle address byte
	spi_transmit((uint8_t)(ftAddress));		// Send low address byte
	spi_transmit((uint8_t)(command));		// Send data low byte
	spi_transmit((uint8_t)(command >> 8));
	spi_transmit((uint8_t)(command >> 16));
	spi_transmit((uint8_t)(command >> 24));		// Send data high byte
	ft800_inc_cmdoffset(4);			// update the command-ram pointer
}


// virtually the same as ft800memWrite32() but calculating the address and offset as the other commands
void ft800_cmd_dl(uint32_t command)
{
	ft800_start_cmd(command);
	ft800_cs_clear();
}


// Write a string to co-processor memory in context of a command: no chip-select, just plain spi-transfers
void ft800_write_string(const char *text)
{
	uint8_t textindex = 0;
	uint8_t padding = 0;

	while(text[textindex] != 0)
	{
		spi_transmit(text[textindex]);
		textindex++;
	}

	padding = textindex % 4;  // 0, 1, 2 oder 3
	padding = 4-padding; // 4, 3, 2, 1

	while(padding > 0)
	{
		spi_transmit(0);
		padding--;
		textindex++;
	}

	ft800_inc_cmdoffset(textindex);	// update the command-ram pointer
}

// commands to draw graphics objects:

void ft800_cmd_text(int16_t x0, int16_t y0, int16_t font, uint16_t options, const char* text)
{
	ft800_start_cmd(CMD_TEXT);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(font));
	spi_transmit((uint8_t)(font >> 8));

	spi_transmit((uint8_t)(options));
	spi_transmit((uint8_t)(options >> 8));

	ft800_inc_cmdoffset(8);	// update the command-ram pointer
	ft800_write_string(text);

	ft800_cs_clear();
}


void ft800_cmd_button(int16_t x0, int16_t y0, int16_t w0, int16_t h0, int16_t font, uint16_t options, const char* text)
{
	ft800_start_cmd(CMD_BUTTON);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(w0));
	spi_transmit((uint8_t)(w0 >> 8));

	spi_transmit((uint8_t)(h0));
	spi_transmit((uint8_t)(h0 >> 8));

	spi_transmit((uint8_t)(font));
	spi_transmit((uint8_t)(font >> 8));

	spi_transmit((uint8_t)(options));
	spi_transmit((uint8_t)(options >> 8));

	ft800_inc_cmdoffset(12);	// update the command-ram pointer
	ft800_write_string(text);
	ft800_cs_clear();
}


// draw a clock, milliseconds are always sent as 0x00
void ft800_cmd_clock(int16_t x0, int16_t y0, int16_t r0, uint16_t options, uint16_t hours, uint16_t minutes, uint16_t seconds)
{
	ft800_start_cmd(CMD_CLOCK);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(r0));
	spi_transmit((uint8_t)(r0 >> 8));

	spi_transmit((uint8_t)(options));
	spi_transmit((uint8_t)(options >> 8));

	spi_transmit((uint8_t)(hours));
	spi_transmit((uint8_t)(hours >> 8));

	spi_transmit((uint8_t)(minutes));
	spi_transmit((uint8_t)(minutes >> 8));

	spi_transmit((uint8_t)(seconds));
	spi_transmit((uint8_t)(seconds >> 8));
	spi_transmit(0x00); // millis low
	spi_transmit(0x00); // millis high

	ft800_cs_clear();
	ft800_inc_cmdoffset(16);	// update the command-ram pointer
}


void ft800_cmd_bgcolor(uint32_t color)
{
	ft800_start_cmd(CMD_BGCOLOR);

	spi_transmit((uint8_t)(color));		// Send data low byte
	spi_transmit((uint8_t)(color >> 8));
	spi_transmit((uint8_t)(color >> 16));
	spi_transmit(0x00);

	ft800_cs_clear();
	ft800_inc_cmdoffset(4);	// update the command-ram pointer
}


void ft800_cmd_fgcolor(uint32_t color)
{
	ft800_start_cmd(CMD_FGCOLOR);

	spi_transmit((uint8_t)(color));		// Send data low byte
	spi_transmit((uint8_t)(color >> 8));
	spi_transmit((uint8_t)(color >> 16));
	spi_transmit(0x00);

	ft800_cs_clear();
	ft800_inc_cmdoffset(4);	// update the command-ram pointer
}


void ft800_cmd_gradcolor(uint32_t color)
{
	ft800_start_cmd(CMD_GRADCOLOR);

	spi_transmit((uint8_t)(color));		// Send data low byte
	spi_transmit((uint8_t)(color >> 8));
	spi_transmit((uint8_t)(color >> 16));
	spi_transmit(0x00);

	ft800_cs_clear();
	ft800_inc_cmdoffset(4);	// update the command-ram pointer
}


void ft800_cmd_gauge(int16_t x0, int16_t y0, int16_t r0, uint16_t options, uint16_t major, uint16_t minor, uint16_t val, uint16_t range)
{
	ft800_start_cmd(CMD_GAUGE);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(r0));
	spi_transmit((uint8_t)(r0 >> 8));

	spi_transmit((uint8_t)(options));
	spi_transmit((uint8_t)(options >> 8));

	spi_transmit((uint8_t)(major));
	spi_transmit((uint8_t)(major >> 8));

	spi_transmit((uint8_t)(minor));
	spi_transmit((uint8_t)(minor >> 8));

	spi_transmit((uint8_t)(val));
	spi_transmit((uint8_t)(val >> 8));

	spi_transmit((uint8_t)(range));
	spi_transmit((uint8_t)(range >> 8));

	ft800_cs_clear();
	ft800_inc_cmdoffset(16);	// update the command-ram pointer
}


void ft800_cmd_gradient(int16_t x0, int16_t y0, uint32_t rgb0, int16_t x1, int16_t y1, uint32_t rgb1)
{
	ft800_start_cmd(CMD_GRADIENT);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(rgb0));		// Send data low byte
	spi_transmit((uint8_t)(rgb0 >> 8));
	spi_transmit((uint8_t)(rgb0 >> 16));
	spi_transmit(0x00);

	spi_transmit((uint8_t)(x1));
	spi_transmit((uint8_t)(x1 >> 8));

	spi_transmit((uint8_t)(y1));
	spi_transmit((uint8_t)(y1 >> 8));

	spi_transmit((uint8_t)(rgb1));		// Send data low byte
	spi_transmit((uint8_t)(rgb1 >> 8));
	spi_transmit((uint8_t)(rgb1 >> 16));
	spi_transmit(0x00);

	ft800_cs_clear();
	ft800_inc_cmdoffset(16);	// update the command-ram pointer
}


void ft800_cmd_keys(int16_t x0, int16_t y0, int16_t w0, int16_t h0, int16_t font, uint16_t options, const char* text)
{
	ft800_start_cmd(CMD_KEYS);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(w0));
	spi_transmit((uint8_t)(w0 >> 8));

	spi_transmit((uint8_t)(h0));
	spi_transmit((uint8_t)(h0 >> 8));

	spi_transmit((uint8_t)(font));
	spi_transmit((uint8_t)(font >> 8));

	spi_transmit((uint8_t)(options));
	spi_transmit((uint8_t)(options >> 8));

	ft800_inc_cmdoffset(12);	// update the command-ram pointer
	ft800_write_string(text);
	ft800_cs_clear();
}


void ft800_cmd_progress(int16_t x0, int16_t y0, int16_t w0, int16_t h0, uint16_t options, uint16_t val, uint16_t range)
{
	ft800_start_cmd(CMD_PROGRESS);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(w0));
	spi_transmit((uint8_t)(w0 >> 8));

	spi_transmit((uint8_t)(h0));
	spi_transmit((uint8_t)(h0 >> 8));

	spi_transmit((uint8_t)(options));
	spi_transmit((uint8_t)(options >> 8));

	spi_transmit((uint8_t)(val));
	spi_transmit((uint8_t)(val >> 8));

	spi_transmit((uint8_t)(range));
	spi_transmit((uint8_t)(range >> 8));

	spi_transmit(0x00); // dummy byte for 4-byte alignement
	spi_transmit(0x00); // dummy byte for 4-byte alignement

	ft800_cs_clear();
	ft800_inc_cmdoffset(16);	// update the command-ram pointer
}


void ft800_cmd_scrollbar(int16_t x0, int16_t y0, int16_t w0, int16_t h0, uint16_t options, uint16_t val, uint16_t size, uint16_t range)
{
	ft800_start_cmd(CMD_SCROLLBAR);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(w0));
	spi_transmit((uint8_t)(w0 >> 8));

	spi_transmit((uint8_t)(h0));
	spi_transmit((uint8_t)(h0 >> 8));

	spi_transmit((uint8_t)(options));
	spi_transmit((uint8_t)(options >> 8));

	spi_transmit((uint8_t)(val));
	spi_transmit((uint8_t)(val >> 8));

	spi_transmit((uint8_t)(size));
	spi_transmit((uint8_t)(size >> 8));

	spi_transmit((uint8_t)(range));
	spi_transmit((uint8_t)(range >> 8));

	ft800_cs_clear();
	ft800_inc_cmdoffset(16);	// update the command-ram pointer
}


void ft800_cmd_slider(int16_t x1, int16_t y1, int16_t w1, int16_t h1, uint16_t options, uint16_t val, uint16_t range)
{
	ft800_start_cmd(CMD_SLIDER);

	spi_transmit((uint8_t)(x1));
	spi_transmit((uint8_t)(x1 >> 8));

	spi_transmit((uint8_t)(y1));
	spi_transmit((uint8_t)(y1 >> 8));

	spi_transmit((uint8_t)(w1));
	spi_transmit((uint8_t)(w1 >> 8));

	spi_transmit((uint8_t)(h1));
	spi_transmit((uint8_t)(h1 >> 8));

	spi_transmit((uint8_t)(options));
	spi_transmit((uint8_t)(options >> 8));

	spi_transmit((uint8_t)(val));
	spi_transmit((uint8_t)(val >> 8));

	spi_transmit((uint8_t)(range));
	spi_transmit((uint8_t)(range >> 8));

	spi_transmit(0x00); // dummy byte for 4-byte alignement
	spi_transmit(0x00); // dummy byte for 4-byte alignement

	ft800_cs_clear();
	ft800_inc_cmdoffset(16);	// update the command-ram pointer
}


void ft800_cmd_dial(int16_t x0, int16_t y0, int16_t r0, uint16_t options, uint16_t val)
{
	ft800_start_cmd(CMD_DIAL);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(r0));
	spi_transmit((uint8_t)(r0 >> 8));

	spi_transmit((uint8_t)(options));
	spi_transmit((uint8_t)(options >> 8));

	spi_transmit((uint8_t)(val));
	spi_transmit((uint8_t)(val >> 8));

	spi_transmit(0);
	spi_transmit(0);

	ft800_cs_clear();
	ft800_inc_cmdoffset(12);	// update the command-ram pointer
}


void ft800_cmd_toggle(int16_t x0, int16_t y0, int16_t w0, int16_t font, uint16_t options, uint16_t state, const char* text)
{
	ft800_start_cmd(CMD_TOGGLE);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(w0));
	spi_transmit((uint8_t)(w0 >> 8));

	spi_transmit((uint8_t)(font));
	spi_transmit((uint8_t)(font >> 8));

	spi_transmit((uint8_t)(options));
	spi_transmit((uint8_t)(options >> 8));

	spi_transmit((uint8_t)(state));
	spi_transmit((uint8_t)(state >> 8));

	ft800_inc_cmdoffset(12);	// update the command-ram pointer
	ft800_write_string(text);
	ft800_cs_clear();
}


#ifdef FT_81X_ENABLE
void ft800_cmd_setbase(uint32_t base)
{
	ft800_start_cmd(CMD_SETBASE);

	spi_transmit((uint8_t)(base));		// Send data low byte
	spi_transmit((uint8_t)(base >> 8));
	spi_transmit((uint8_t)(base >> 16));
	spi_transmit((uint8_t)(base >> 24));		// Send data high byte

	ft800_cs_clear();

	ft800_inc_cmdoffset(12);	// update the command-ram pointer	
}
#endif

#ifdef FT_81X_ENABLE
void ft800_cmd_setbitmap(uint32_t addr, uint16_t fmt, uint16_t width, uint16_t height)
{
	ft800_start_cmd(CMD_SETBITMAP);

	spi_transmit((uint8_t)(addr));
	spi_transmit((uint8_t)(addr >> 8));
	spi_transmit((uint8_t)(addr >> 16));
	spi_transmit((uint8_t)(addr >> 24));
	
	spi_transmit((uint8_t)(fmt));
	spi_transmit((uint8_t)(fmt>> 8));

	spi_transmit((uint8_t)(width));
	spi_transmit((uint8_t)(width >> 8));

	spi_transmit((uint8_t)(height));
	spi_transmit((uint8_t)(height >> 8));

	spi_transmit(0);
	spi_transmit(0);	
	
	ft800_cs_clear();
	ft800_inc_cmdoffset(12);	// update the command-ram pointer	
}
#endif


void ft800_cmd_number(int16_t x0, int16_t y0, int16_t font, uint16_t options, int32_t number)
{
	ft800_start_cmd(CMD_NUMBER);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(font));
	spi_transmit((uint8_t)(font >> 8));

	spi_transmit((uint8_t)(options));
	spi_transmit((uint8_t)(options >> 8));

	spi_transmit((uint8_t)(number));		// Send data low byte
	spi_transmit((uint8_t)(number >> 8));
	spi_transmit((uint8_t)(number >> 16));
	spi_transmit((uint8_t)(number >> 24));		// Send data high byte

	ft800_cs_clear();

	ft800_inc_cmdoffset(12);	// update the command-ram pointer
}


// commands to operate on memory:

void ft800_cmd_memzero(uint32_t ptr, uint32_t num)
{
	ft800_start_cmd(CMD_MEMZERO);

	spi_transmit((uint8_t)(ptr));
	spi_transmit((uint8_t)(ptr >> 8));
	spi_transmit((uint8_t)(ptr >> 16));
	spi_transmit((uint8_t)(ptr >> 24));

	spi_transmit((uint8_t)(num));
	spi_transmit((uint8_t)(num >> 8));
	spi_transmit((uint8_t)(num >> 16));
	spi_transmit((uint8_t)(num >> 24));

	ft800_cs_clear();
	ft800_inc_cmdoffset(8);	// update the command-ram pointer
}


void ft800_cmd_memset(uint32_t ptr, uint8_t value, uint32_t num)
{
	ft800_start_cmd(CMD_MEMSET);

	spi_transmit((uint8_t)(ptr));
	spi_transmit((uint8_t)(ptr >> 8));
	spi_transmit((uint8_t)(ptr >> 16));
	spi_transmit((uint8_t)(ptr >> 24));

	spi_transmit(value);
	spi_transmit(0);
	spi_transmit(0);
	spi_transmit(0);

	spi_transmit((uint8_t)(num));
	spi_transmit((uint8_t)(num >> 8));
	spi_transmit((uint8_t)(num >> 16));
	spi_transmit((uint8_t)(num >> 24));

	ft800_cs_clear();
	ft800_inc_cmdoffset(12);	// update the command-ram pointer
}


/*
void ft800_cmd_memwrite(uint32_t dest, uint32_t num, const uint8_t *data)
{
	ft800_start_cmd(CMD_MEMWRITE);

	spi_transmit((uint8_t)(dest));
	spi_transmit((uint8_t)(dest >> 8));
	spi_transmit((uint8_t)(dest >> 16));
	spi_transmit((uint8_t)(dest >> 24));

	spi_transmit((uint8_t)(num));
	spi_transmit((uint8_t)(num >> 8));
	spi_transmit((uint8_t)(num >> 16));
	spi_transmit((uint8_t)(num >> 24));

	num = (num + 3)&(~3);

	for(count=0;count<len;count++)
	{
		spi_transmit(pgm_read_byte_far(data+count));
	}

	ft800_cs_clear();
	ft800_inc_cmdoffset(8+len);	// update the command-ram pointer
}
*/


void ft800_cmd_memcpy(uint32_t dest, uint32_t src, uint32_t num)
{
	ft800_start_cmd(CMD_MEMCPY);

	spi_transmit((uint8_t)(dest));
	spi_transmit((uint8_t)(dest >> 8));
	spi_transmit((uint8_t)(dest >> 16));
	spi_transmit((uint8_t)(dest >> 24));

	spi_transmit((uint8_t)(src));
	spi_transmit((uint8_t)(src >> 8));
	spi_transmit((uint8_t)(src >> 16));
	spi_transmit((uint8_t)(src >> 24));

	spi_transmit((uint8_t)(num));
	spi_transmit((uint8_t)(num >> 8));
	spi_transmit((uint8_t)(num >> 16));
	spi_transmit((uint8_t)(num >> 24));

	ft800_cs_clear();
	ft800_inc_cmdoffset(12);	// update the command-ram pointer
}


void ft800_cmd_append(uint32_t ptr, uint32_t num)
{
	ft800_start_cmd(CMD_APPEND);

	spi_transmit((uint8_t)(ptr));
	spi_transmit((uint8_t)(ptr >> 8));
	spi_transmit((uint8_t)(ptr >> 16));
	spi_transmit((uint8_t)(ptr >> 24));

	spi_transmit((uint8_t)(num));
	spi_transmit((uint8_t)(num >> 8));
	spi_transmit((uint8_t)(num >> 16));
	spi_transmit((uint8_t)(num >> 24));

	ft800_cs_clear();
	ft800_inc_cmdoffset(8);	// update the command-ram pointer
}


// commands for loading image data into FT800 memory:

void spi_flash_write(const uint8_t *data, uint16_t len)
{
	uint16_t count;

	len = (len + 3)&(~3);

	for(count=0;count<len;count++)
	{
		spi_transmit(fetch_flash_byte(data+count));
	}

	ft800_inc_cmdoffset(len);	// update the command-ram pointer
}


void ft800_cmd_inflate(uint32_t ptr, const uint8_t *data, uint16_t len)
{
	ft800_start_cmd(CMD_INFLATE);

	spi_transmit((uint8_t)(ptr));		// Send data low byte
	spi_transmit((uint8_t)(ptr >> 8));
	spi_transmit((uint8_t)(ptr >> 16));
	spi_transmit((uint8_t)(ptr >> 24));		// Send data high byte

	ft800_inc_cmdoffset(4);	// update the command-ram pointer

	spi_flash_write(data,len);

	ft800_cs_clear();
}


void ft800_cmd_loadimage(uint32_t ptr, uint32_t options, const uint8_t *data, uint16_t len)
{
	ft800_start_cmd(CMD_LOADIMAGE);

	spi_transmit((uint8_t)(ptr));		// Send data low byte
	spi_transmit((uint8_t)(ptr >> 8));
	spi_transmit((uint8_t)(ptr >> 16));
	spi_transmit((uint8_t)(ptr >> 24));		// Send data high byte

	spi_transmit((uint8_t)(options));		// Send data low byte
	spi_transmit((uint8_t)(options >> 8));
	spi_transmit((uint8_t)(options >> 16));
	spi_transmit((uint8_t)(options >> 24));		// Send data high byte

	ft800_inc_cmdoffset(8);	// update the command-ram pointer

	spi_flash_write(data,len);

	ft800_cs_clear();
}


#ifdef FT_81X_ENABLE
void ft800_cmd_mediafifo(uint32_t ptr, uint32_t size)
{
	ft800_start_cmd(CMD_MEDIAFIFO);

	spi_transmit((uint8_t)(ptr));		// Send data low byte
	spi_transmit((uint8_t)(ptr >> 8));
	spi_transmit((uint8_t)(ptr >> 16));
	spi_transmit((uint8_t)(ptr >> 24));		// Send data high byte

	spi_transmit((uint8_t)(size));		// Send data low byte
	spi_transmit((uint8_t)(size >> 8));
	spi_transmit((uint8_t)(size >> 16));
	spi_transmit((uint8_t)(size >> 24));		// Send data high byte

	ft800_cs_clear();

	ft800_inc_cmdoffset(8);	// update the command-ram pointer
}
#endif



// commands for setting the bitmap transform matrix:

void ft800_cmd_loadidentity(void)
{
	ft800_start_cmd(CMD_LOADIDENTITY);
	ft800_cs_clear();
}


void ft800_cmd_translate(int32_t tx, int32_t ty)
{
	ft800_start_cmd(CMD_TRANSLATE);

	spi_transmit((uint8_t)(tx));		// Send data low byte
	spi_transmit((uint8_t)(tx >> 8));
	spi_transmit((uint8_t)(tx >> 16));
	spi_transmit((uint8_t)(tx >> 24));		// Send data high byte

	spi_transmit((uint8_t)(ty));		// Send data low byte
	spi_transmit((uint8_t)(ty >> 8));
	spi_transmit((uint8_t)(ty >> 16));
	spi_transmit((uint8_t)(ty >> 24));		// Send data high byte

	ft800_cs_clear();
	ft800_inc_cmdoffset(8);	// update the command-ram pointer
}


void ft800_cmd_scale(int32_t sx, int32_t sy)
{
	ft800_start_cmd(CMD_SCALE);

	spi_transmit((uint8_t)(sx));		// Send data low byte
	spi_transmit((uint8_t)(sx >> 8));
	spi_transmit((uint8_t)(sx >> 16));
	spi_transmit((uint8_t)(sx >> 24));		// Send data high byte

	spi_transmit((uint8_t)(sy));		// Send data low byte
	spi_transmit((uint8_t)(sy >> 8));
	spi_transmit((uint8_t)(sy >> 16));
	spi_transmit((uint8_t)(sy >> 24));		// Send data high byte

	ft800_cs_clear();
	ft800_inc_cmdoffset(8);	// update the command-ram pointer
}


void ft800_cmd_rotate(int32_t ang)
{
	ft800_start_cmd(CMD_ROTATE);

	spi_transmit((uint8_t)(ang));		// Send data low byte
	spi_transmit((uint8_t)(ang >> 8));
	spi_transmit((uint8_t)(ang >> 16));
	spi_transmit((uint8_t)(ang >> 24));		// Send data high byte

	ft800_cs_clear();
	ft800_inc_cmdoffset(4);	// update the command-ram pointer
}


void ft800_cmd_setmatrix(void)
{
	ft800_start_cmd(CMD_SETMATRIX);
	ft800_cs_clear();
}


// the description in the programmers guide is strange for this funtion
// while it is named *get*matrix, parameters 'a' to 'f' are supplied to
// the function and described as "output parameter"
// best guess is that this one allows to setup the matrix coefficients manually
// instead automagically like with _translate, _scale and _rotate
// if this assumption is correct it rather should be named cmd_setupmatrix()
void ft800_cmd_getmatrix(int32_t a, int32_t b, int32_t c, int32_t d, int32_t e, int32_t f)
{
	ft800_start_cmd(CMD_SETMATRIX);

	spi_transmit((uint8_t)(a));
	spi_transmit((uint8_t)(a >> 8));
	spi_transmit((uint8_t)(a >> 16));
	spi_transmit((uint8_t)(a >> 24));

	spi_transmit((uint8_t)(b));
	spi_transmit((uint8_t)(b >> 8));
	spi_transmit((uint8_t)(b >> 16));
	spi_transmit((uint8_t)(b >> 24));

	spi_transmit((uint8_t)(c));
	spi_transmit((uint8_t)(c >> 8));
	spi_transmit((uint8_t)(c >> 16));
	spi_transmit((uint8_t)(c >> 24));

	spi_transmit((uint8_t)(d));
	spi_transmit((uint8_t)(d >> 8));
	spi_transmit((uint8_t)(d >> 16));
	spi_transmit((uint8_t)(d >> 24));

	spi_transmit((uint8_t)(e));
	spi_transmit((uint8_t)(e >> 8));
	spi_transmit((uint8_t)(e >> 16));
	spi_transmit((uint8_t)(e >> 24));

	spi_transmit((uint8_t)(f));
	spi_transmit((uint8_t)(f >> 8));
	spi_transmit((uint8_t)(f >> 16));
	spi_transmit((uint8_t)(f >> 24));

	ft800_cs_clear();
	ft800_inc_cmdoffset(24);	// update the command-ram pointer
}


// other commands:

void ft800_cmd_calibrate(void)
{
	ft800_start_cmd(CMD_CALIBRATE);

	spi_transmit(0);
	spi_transmit(0);
	spi_transmit(0);
	spi_transmit(0);

	ft800_cs_clear();
	ft800_inc_cmdoffset(4);	// update the command-ram pointer
}


void ft800_cmd_coldstart(void)
{
	ft800_start_cmd(CMD_COLDSTART);
	ft800_cs_clear();
}


void ft800_cmd_interrupt(uint32_t ms)
{
	ft800_start_cmd(CMD_SNAPSHOT);

	spi_transmit((uint8_t)(ms));
	spi_transmit((uint8_t)(ms >> 8));
	spi_transmit((uint8_t)(ms >> 16));
	spi_transmit((uint8_t)(ms >> 24));

	ft800_cs_clear();
	ft800_inc_cmdoffset(4);	// update the command-ram pointer
}


void ft800_cmd_logo(void)
{
	ft800_start_cmd(CMD_LOGO);
	ft800_cs_clear();
}


void ft800_cmd_screensaver(void)
{
	ft800_start_cmd(CMD_SCREENSAVER);
	ft800_cs_clear();
}


#ifdef FT_81X_ENABLE
void ft800_cmd_setrotate(uint32_t r)
{
	ft800_start_cmd(CMD_SETROTATE);

	spi_transmit((uint8_t)(r));
	spi_transmit((uint8_t)(r >> 8));
	spi_transmit((uint8_t)(r >> 16));
	spi_transmit((uint8_t)(r >> 24));

	ft800_cs_clear();
	ft800_inc_cmdoffset(4);	// update the command-ram pointer
}
#endif


void ft800_cmd_sketch(int16_t x0, int16_t y0, uint16_t w0, uint16_t h0, uint32_t ptr, uint16_t format)
{
	ft800_start_cmd(CMD_SKETCH);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(w0));
	spi_transmit((uint8_t)(w0 >> 8));

	spi_transmit((uint8_t)(h0));
	spi_transmit((uint8_t)(h0 >> 8));

	spi_transmit((uint8_t)(ptr));
	spi_transmit((uint8_t)(ptr >> 8));
	spi_transmit((uint8_t)(ptr >> 16));
	spi_transmit((uint8_t)(ptr >> 24));

	spi_transmit((uint8_t)(format));
	spi_transmit((uint8_t)(format >> 8));

	spi_transmit(0);
	spi_transmit(0);

	ft800_cs_clear();
	ft800_inc_cmdoffset(16);	// update the command-ram pointer
}


void ft800_cmd_snapshot(uint32_t ptr)
{
	ft800_start_cmd(CMD_SNAPSHOT);

	spi_transmit((uint8_t)(ptr));
	spi_transmit((uint8_t)(ptr >> 8));
	spi_transmit((uint8_t)(ptr >> 16));
	spi_transmit((uint8_t)(ptr >> 24));

	ft800_cs_clear();
	ft800_inc_cmdoffset(4);	// update the command-ram pointer
}


void ft800_cmd_spinner(int16_t x0, int16_t y0, uint16_t style, uint16_t scale)
{
	ft800_start_cmd(CMD_SPINNER);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(style));
	spi_transmit((uint8_t)(style >> 8));

	spi_transmit((uint8_t)(scale));
	spi_transmit((uint8_t)(scale >> 8));

	ft800_cs_clear();
	ft800_inc_cmdoffset(8);	// update the command-ram pointer
}


void ft800_cmd_stop(void)
{
	ft800_start_cmd(CMD_STOP);
	ft800_cs_clear();
}


void ft800_cmd_track(int16_t x0, int16_t y0, int16_t w0, int16_t h0, int16_t tag)
{
	ft800_start_cmd(CMD_TRACK);

	spi_transmit((uint8_t)(x0));
	spi_transmit((uint8_t)(x0 >> 8));

	spi_transmit((uint8_t)(y0));
	spi_transmit((uint8_t)(y0 >> 8));

	spi_transmit((uint8_t)(w0));
	spi_transmit((uint8_t)(w0 >> 8));

	spi_transmit((uint8_t)(h0));
	spi_transmit((uint8_t)(h0 >> 8));

	spi_transmit((uint8_t)(tag));
	spi_transmit((uint8_t)(tag >> 8));

	spi_transmit(0);
	spi_transmit(0);

	ft800_cs_clear();
	ft800_inc_cmdoffset(12);	// update the command-ram pointer
}


// meta-commands, sequences of several display-list entries condensed into simpler to use functions

void ft800_cmd_point(int16_t x0, int16_t y0, uint16_t size)
{
	uint32_t calc;

	ft800_start_cmd((DL_BEGIN | FT_POINTS));

	calc = POINT_SIZE(size*16);
	spi_transmit((uint8_t)(calc));
	spi_transmit((uint8_t)(calc >> 8));
	spi_transmit((uint8_t)(calc >> 16));
	spi_transmit((uint8_t)(calc >> 24));

	calc = VERTEX2F(x0 * 16, y0 * 16);
	spi_transmit((uint8_t)(calc));
	spi_transmit((uint8_t)(calc >> 8));
	spi_transmit((uint8_t)(calc >> 16));
	spi_transmit((uint8_t)(calc >> 24));

	spi_transmit((uint8_t)(DL_END));		// Send data low byte
	spi_transmit((uint8_t)(DL_END >> 8));
	spi_transmit((uint8_t)(DL_END >> 16));
	spi_transmit((uint8_t)(DL_END >> 24));		// Send data high byte

	ft800_cs_clear();
	ft800_inc_cmdoffset(12);	// update the command-ram pointer
}


void ft800_cmd_line(int16_t x0, int16_t y0, int16_t x1, int16_t y1, uint16_t width)
{
	uint32_t calc;

	ft800_start_cmd((DL_BEGIN | FT_LINES));

	calc = LINE_WIDTH(width * 16);
	spi_transmit((uint8_t)(calc));
	spi_transmit((uint8_t)(calc >> 8));
	spi_transmit((uint8_t)(calc >> 16));
	spi_transmit((uint8_t)(calc >> 24));

	calc = VERTEX2F(x0 * 16, y0 * 16);
	spi_transmit((uint8_t)(calc));
	spi_transmit((uint8_t)(calc >> 8));
	spi_transmit((uint8_t)(calc >> 16));
	spi_transmit((uint8_t)(calc >> 24));

	calc = VERTEX2F(x1 * 16, y1 * 16);
	spi_transmit((uint8_t)(calc));
	spi_transmit((uint8_t)(calc >> 8));
	spi_transmit((uint8_t)(calc >> 16));
	spi_transmit((uint8_t)(calc >> 24));

	spi_transmit((uint8_t)(DL_END));		// Send data low byte
	spi_transmit((uint8_t)(DL_END >> 8));
	spi_transmit((uint8_t)(DL_END >> 16));
	spi_transmit((uint8_t)(DL_END >> 24));		// Send data high byte

	ft800_cs_clear();
	ft800_inc_cmdoffset(16);	// update the command-ram pointer
}


void ft800_cmd_rect(int16_t x0, int16_t y0, int16_t x1, int16_t y1, uint16_t corner)
{
	uint32_t calc;

	ft800_start_cmd((DL_BEGIN | FT_RECTS));

	calc = LINE_WIDTH(corner * 16);
	spi_transmit((uint8_t)(calc));
	spi_transmit((uint8_t)(calc >> 8));
	spi_transmit((uint8_t)(calc >> 16));
	spi_transmit((uint8_t)(calc >> 24));

	calc = VERTEX2F(x0 * 16, y0 * 16);
	spi_transmit((uint8_t)(calc));
	spi_transmit((uint8_t)(calc >> 8));
	spi_transmit((uint8_t)(calc >> 16));
	spi_transmit((uint8_t)(calc >> 24));

	calc = VERTEX2F(x1 * 16, y1 * 16);
	spi_transmit((uint8_t)(calc));
	spi_transmit((uint8_t)(calc >> 8));
	spi_transmit((uint8_t)(calc >> 16));
	spi_transmit((uint8_t)(calc >> 24));

	spi_transmit((uint8_t)(DL_END));		// Send data low byte
	spi_transmit((uint8_t)(DL_END >> 8));
	spi_transmit((uint8_t)(DL_END >> 16));
	spi_transmit((uint8_t)(DL_END >> 24));		// Send data high byte

	ft800_cs_clear();
	ft800_inc_cmdoffset(16);	// update the command-ram pointer
}


// init

void ft800_init(void)
{
	uint8_t gpio;
	uint8_t chipid;

	ft800_pdn_set();
	DELAY_MS(6);	// minimum time for power-down is 5ms
	ft800_pdn_clear();
	DELAY_MS(21);	// minimum time to allow from rising PD_N to first access is 20ms

//	ft800_cmdWrite(FT_CLKEXT);	// Set FT800 for external clock
	ft800_cmdWrite(FT_CLKINT);	// Set FT800 for internal clock
// 	ft800_cmdWrite(FT_CLK48M);	// Set FT800 for 48MHz PLL

	ft800_cmdWrite(FT_ACTIVE);	// Start FT800

	chipid = ft800_memRead8(REG_ID);	// Read ID register
	while(chipid != 0x7C)	// if chipid is not 0x7c, continue to read it until it is, FT81x may need a moment for it's power on selftest
	{
		chipid = ft800_memRead8(REG_ID);
	}

	ft800_memWrite8(REG_PCLK, 0x00);		// Set PCLK to zero - don't clock the LCD until later
	ft800_memWrite8(REG_PWM_DUTY, 10);		// Turn off backlight

	// Initialize Display
	ft800_memWrite16(REG_HSIZE,   FT_HSIZE);	// active display width
	ft800_memWrite16(REG_HCYCLE,  FT_HCYCLE);	// total number of clocks per line, incl front/back porch
	ft800_memWrite16(REG_HOFFSET, FT_HOFFSET);	// start of active line
	ft800_memWrite16(REG_HSYNC0,  FT_HSYNC0);	// start of horizontal sync pulse
	ft800_memWrite16(REG_HSYNC1,  FT_HSYNC1);	// end of horizontal sync pulse
	ft800_memWrite16(REG_VSIZE,   FT_VSIZE);	// active display height
	ft800_memWrite16(REG_VCYCLE,  FT_VCYCLE);	// total number of lines per screen, incl pre/post
	ft800_memWrite16(REG_VOFFSET, FT_VOFFSET);	// start of active screen
	ft800_memWrite16(REG_VSYNC0,  FT_VSYNC0);	// start of vertical sync pulse
	ft800_memWrite16(REG_VSYNC1,  FT_VSYNC1);	// end of vertical sync pulse
	ft800_memWrite8(REG_SWIZZLE,  FT_SWIZZLE);	// FT800 output to LCD - pin order
	ft800_memWrite8(REG_PCLK_POL, FT_PCLKPOL);	// LCD data is clocked in on this PCLK edge
	// Don't set PCLK yet - wait for just after the first display list

	// Configure Touch
	ft800_memWrite8(REG_TOUCH_MODE, FT_TMODE_CONTINUOUS);	// enable touch
	ft800_memWrite16(REG_TOUCH_RZTHRESH, FT_TOUCH_RZTHRESH);	// Eliminate any false touches

	// Configure Audio - not used, so disable it
	ft800_memWrite8(REG_VOL_PB, 0x00);		// turn recorded audio volume down
//	ft800_memWrite8(REG_VOL_SOUND, 0xff);		// turn synthesizer volume on
	ft800_memWrite8(REG_VOL_SOUND, 0x00);		// turn synthesizer volume off
	ft800_memWrite16(REG_SOUND, 0x6000);		// set synthesizer to mute

	ft800_memWrite32(FT_RAM_DL, DL_CLEAR_RGB);
	ft800_memWrite32(FT_RAM_DL + 4, (DL_CLEAR | CLR_COL | CLR_STN | CLR_TAG));
	ft800_memWrite32(FT_RAM_DL + 8, DL_DISPLAY);	// end of display list
	ft800_memWrite32(REG_DLSWAP, FT_DLSWAP_FRAME);

	// Nothing is being displayed yet... the pixel clock is still 0x00

	gpio = ft800_memRead8(REG_GPIO_DIR);
	gpio |= 0x82; // set DISP to Output although it always is output, set GPIO1 to Output - Audio Enable on VM800B
	ft800_memWrite8(REG_GPIO_DIR, gpio);

	gpio = ft800_memRead8(REG_GPIO);	// Read the FT800 GPIO register for a read/modify/write operation
//	gpio |= 0x82;						// set bit 7 of FT800 GPIO register (DISP), set GPIO1 to High to enable Audio - others are inputs
	gpio |= 0x80;						// set bit 7 of FT800 GPIO register (DISP), others are inputs
	ft800_memWrite8(REG_GPIO, gpio);	// Enable the DISP signal to the LCD panel
	ft800_memWrite8(REG_PCLK, FT_PCLK);	// Now start clocking data to the LCD panel

	ft800_memWrite8(REG_PWM_DUTY, 80);	// Turn on backlight

	DELAY_MS(10);	// just to be safe
	
	while(ft800_busy() == 1);
	ft800_get_cmdoffset();
}
